import threading
from binascii import b2a_hex


def main():
    num_threads = 10
    use_threads = True

    if not use_threads:
        # Run core code
        runShapelyBuilding()
    else:
        threads = [
            threading.Thread(target=runShapelyBuilding, name=str(i), args=(i,))
            for i in range(num_threads)
        ]
        for t in threads:
            t.start()
        for t in threads:
            t.join()


def runShapelyBuilding(num):
    print(f"{num}: Running shapely tests on wkb")
    import shapely.geos

    print(f"{num} GEOS Handle: {shapely.geos.lgeos.geos_handle}")
    import shapely.wkb
    import shapely.wkt

    p = shapely.wkt.loads("POINT (0 0)")
    print(f"{num} WKT: {shapely.wkt.dumps(p)}")
    wkb = shapely.wkb.dumps(p)
    print(f"{num} WKB: {b2a_hex(wkb)}")

    for i in range(10):
        shapely.wkb.loads(wkb)

    print(f"{num} GEOS Handle: {shapely.geos.lgeos.geos_handle}")
    print(f"Done {num}")


if __name__ == "__main__":
    main()



import numpy as np
import pytest

import shapely
from shapely import (
    Geometry,
    LineString,
    MultiPolygon,
    Polygon,
)
from shapely.errors import UnsupportedGEOSVersionError
from shapely.testing import assert_geometries_equal
from shapely.tests.common import (
    all_types,
    all_types_z,
    empty_line_string,
)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="requires >= 3.12")
@pytest.mark.parametrize("geometry", all_types + all_types_z)
def test_coverage_is_valid(geometry):
    actual = shapely.coverage_is_valid([geometry])
    assert actual.ndim == 0
    assert actual.dtype == np.bool_
    assert actual.item() is True

    actual = shapely.coverage_invalid_edges([geometry])
    expected = np.array([empty_line_string], dtype=object)
    assert_geometries_equal(actual, expected)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="requires >= 3.12")
def test_coverage_is_valid_non_polygonal():
    # non-polygonal geometries are ignored to validate the coverage
    # (e.g. even if you have crossing linestrings)
    geoms = [
        LineString([(0, 0), (1, 1)]),
        LineString([(1, 0), (0, 1)]),
        Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]),
    ]
    assert shapely.coverage_is_valid(geoms)
    assert (shapely.coverage_invalid_edges(geoms) == empty_line_string).all()


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="requires >= 3.12")
def test_coverage_is_valid_polygonal():
    # adjacent triangles
    poly1 = Polygon([(0, 0), (1, 1), (1, 0), (0, 0)])
    poly2 = Polygon([(0, 0), (1, 1), (0, 1), (0, 0)])
    assert shapely.coverage_is_valid([poly1, poly2])
    assert shapely.is_empty(shapely.coverage_invalid_edges([poly1, poly2])).all()

    # shared egde but without identical vertices
    poly2b = Polygon([(0, 0), (0.5, 0.5), (1, 1), (0, 1), (0, 0)])
    assert not shapely.coverage_is_valid([poly1, poly2b])
    result = shapely.coverage_invalid_edges([poly1, poly2b])
    expected = [LineString([(0, 0), (1, 1)]), LineString([(0, 0), (0.5, 0.5), (1, 1)])]
    assert_geometries_equal(result, expected)

    # overlap
    poly3 = Polygon([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)])
    assert not shapely.coverage_is_valid([poly1, poly3])
    result = shapely.coverage_invalid_edges([poly1, poly3])
    assert not shapely.is_empty(result).any()


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="requires >= 3.12")
def test_coverage_is_valid_gap_width():
    # shared edge of boxes with multiple vertices
    poly1 = shapely.from_wkt("POLYGON ((0 10, 10 10, 10 7, 10 3, 10 0, 0 0, 0 10))")
    poly2 = shapely.from_wkt("POLYGON ((10 10, 20 10, 20 0, 10 0, 10 3, 10 7, 10 10))")

    # extra vertex in middle of shared edge
    poly2_extra = shapely.from_wkt(
        "POLYGON ((10 10, 20 10, 20 0, 10 0, 10 3, 10 5, 10 7, 10 10))"
    )
    # extra vertex shifted -> gap of max 1 wide
    poly2_shift = shapely.from_wkt(
        "POLYGON ((10 10, 20 10, 20 0, 10 0, 10 3, 11 5, 10 7, 10 10))"
    )

    # valid coverage -> gap_width value does not matter
    assert shapely.coverage_is_valid([poly1, poly2], gap_width=0.0)
    assert shapely.coverage_is_valid([poly1, poly2], gap_width=2.0)

    result = shapely.coverage_invalid_edges([poly1, poly2], gap_width=0.0)
    assert_geometries_equal(result, [empty_line_string] * 2)
    result = shapely.coverage_invalid_edges([poly1, poly2], gap_width=2.0)
    assert_geometries_equal(result, [empty_line_string] * 2)

    # invalid coverage -> gap_width value does not matter
    assert not shapely.coverage_is_valid([poly1, poly2_extra], gap_width=0.0)
    assert not shapely.coverage_is_valid([poly1, poly2_extra], gap_width=2.0)

    expected = shapely.from_wkt(
        ["LINESTRING (10 7, 10 3)", "LINESTRING (10 3, 10 5, 10 7)"]
    )
    result = shapely.coverage_invalid_edges([poly1, poly2_extra], gap_width=0.0)
    assert_geometries_equal(result, expected)
    result = shapely.coverage_invalid_edges([poly1, poly2_extra], gap_width=2.0)
    assert_geometries_equal(result, expected)

    # coverage with gap of 1 unit wide
    assert shapely.coverage_is_valid([poly1, poly2_shift], gap_width=0.0)
    assert shapely.coverage_is_valid([poly1, poly2_shift], gap_width=0.5)
    assert not shapely.coverage_is_valid([poly1, poly2_shift], gap_width=1.0)
    assert not shapely.coverage_is_valid([poly1, poly2_shift], gap_width=1.5)
    # TODO why this behaviour?
    assert shapely.coverage_is_valid([poly1, poly2_shift], gap_width=2.0)

    assert_geometries_equal(
        shapely.coverage_invalid_edges([poly1, poly2_shift], gap_width=0.0),
        [empty_line_string] * 2,
    )
    assert_geometries_equal(
        shapely.coverage_invalid_edges([poly1, poly2_shift], gap_width=1.0),
        shapely.from_wkt(["LINESTRING (10 7, 10 3)", "LINESTRING (10 3, 11 5, 10 7)"]),
    )


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="requires >= 3.12")
def test_coverage_invalid_edges_gufunc():
    poly1 = shapely.from_wkt("POLYGON ((0 10, 10 10, 10 7, 10 3, 10 0, 0 0, 0 10))")
    poly2 = shapely.from_wkt("POLYGON ((10 10, 20 10, 20 0, 10 0, 10 3, 10 7, 10 10))")
    poly2_extra = shapely.from_wkt(
        "POLYGON ((10 10, 20 10, 20 0, 10 0, 10 3, 10 5, 10 7, 10 10))"
    )
    poly3 = shapely.from_wkt("POLYGON ((20 10, 30 10, 30 7, 30 3, 30 0, 20 0, 20 10))")

    arr = np.array([[poly1, poly2, poly3], [poly1, poly2_extra, poly3]])
    result = shapely.lib.coverage_invalid_edges(arr, 0.0)
    expected = shapely.from_wkt(
        [
            ["LINESTRING EMPTY"] * 3,
            [
                "LINESTRING (10 7, 10 3)",
                "LINESTRING (10 3, 10 5, 10 7)",
                "LINESTRING EMPTY",
            ],
        ]
    )
    assert_geometries_equal(result, expected)

    arr2 = np.array(arr, order="F")
    result = shapely.lib.coverage_invalid_edges(arr2, 0.0)
    assert_geometries_equal(result, expected)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
@pytest.mark.parametrize("geometry", all_types)
def test_coverage_simplify_scalars(geometry):
    if geometry.geom_type in ("Polygon", "MultiPolygon"):
        actual = shapely.coverage_simplify(geometry, 0.0)
        assert isinstance(actual, Geometry)
        assert shapely.get_type_id(actual) == shapely.get_type_id(geometry)
        assert actual.equals(geometry)
    else:
        with pytest.raises(TypeError, match="incorrect geometry type"):
            shapely.coverage_simplify(geometry, 0.0)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
@pytest.mark.parametrize("geometry", all_types)
def test_coverage_simplify_geom_types(geometry):
    if geometry.geom_type in ("Polygon", "MultiPolygon"):
        actual = shapely.coverage_simplify([geometry, geometry], 0.0)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2,)
        assert (shapely.get_type_id(actual) == shapely.get_type_id(geometry)).all()
    else:
        with pytest.raises(TypeError, match="incorrect geometry type"):
            shapely.coverage_simplify([geometry, geometry], 0.0)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
def test_coverage_simplify_multipolygon():
    mp = MultiPolygon(
        [
            Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]),
            Polygon([(2, 2), (2, 3), (3, 3), (3, 2), (2, 2)]),
        ]
    )
    actual = shapely.coverage_simplify(mp, 1)
    assert actual.equals(
        shapely.from_wkt(
            "MULTIPOLYGON (((0 1, 1 1, 1 0, 0 1)), ((2 3, 3 3, 3 2, 2 3)))"
        )
    )


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
def test_coverage_simplify_array():
    polygons = np.array(
        [
            shapely.Polygon([(0, 0), (20, 0), (20, 10), (10, 5), (0, 10), (0, 0)]),
            shapely.Polygon([(0, 10), (10, 5), (20, 10), (20, 20), (0, 20), (0, 10)]),
        ]
    )
    low_tolerance = shapely.coverage_simplify(polygons, 1)
    mid_tolerance = shapely.coverage_simplify(polygons, 8)
    high_tolerance = shapely.coverage_simplify(polygons, 10)

    assert shapely.equals(low_tolerance, shapely.normalize(polygons)).all()
    assert shapely.equals(
        mid_tolerance,
        shapely.from_wkt(
            [
                "POLYGON ((20 10, 0 10, 0 0, 20 0, 20 10))",
                "POLYGON ((20 10, 0 10, 0 20, 20 20, 20 10))",
            ]
        ),
    ).all()
    assert shapely.equals(
        high_tolerance,
        shapely.from_wkt(
            [
                "POLYGON ((20 10, 0 10, 20 0, 20 10))",
                "POLYGON ((20 10, 0 10, 0 20, 20 10))",
            ]
        ),
    ).all()

    no_boundary = shapely.coverage_simplify(polygons, 10, simplify_boundary=False)
    assert shapely.equals(
        no_boundary,
        shapely.from_wkt(
            [
                "POLYGON ((20 10, 0 10, 0 0, 20 0, 20 10))",
                "POLYGON ((20 10, 0 10, 0 20, 20 20, 20 10))",
            ]
        ),
    ).all()


@pytest.mark.skipif(shapely.geos_version >= (3, 12, 0), reason="requires >= 3.12")
def test_coverage_unsupported_geos():
    geoms = [
        Polygon([(0, 0), (1, 1), (1, 0), (0, 0)]),
        Polygon([(0, 0), (1, 1), (0, 1), (0, 0)]),
    ]
    with pytest.raises(UnsupportedGEOSVersionError):
        shapely.coverage_is_valid(geoms)

    with pytest.raises(UnsupportedGEOSVersionError):
        shapely.coverage_invalid_edges(geoms)

    with pytest.raises(UnsupportedGEOSVersionError):
        shapely.coverage_simplify(geoms, 1.0)



"""Set-theoretic operations on geometry objects."""

import warnings

import numpy as np

from shapely import Geometry, GeometryType, lib
from shapely.decorators import (
    deprecate_positional,
    multithreading_enabled,
    requires_geos,
)

__all__ = [
    "coverage_union",
    "coverage_union_all",
    "difference",
    "disjoint_subset_union",
    "disjoint_subset_union_all",
    "intersection",
    "intersection_all",
    "symmetric_difference",
    "symmetric_difference_all",
    "unary_union",
    "union",
    "union_all",
]


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   difference(a, b, grid_size=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'grid_size' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'b'
#   difference(a, b, *, grid_size=None, **kwargs)


@deprecate_positional(["grid_size"], category=DeprecationWarning)
@multithreading_enabled
def difference(a, b, grid_size=None, **kwargs):
    """Return the part of geometry A that does not intersect with geometry B.

    If grid_size is nonzero, input coordinates will be snapped to a precision
    grid of that size and resulting coordinates will be snapped to that same
    grid.  If 0, this operation will use double precision coordinates.  If None,
    the highest precision of the inputs will be used, which may be previously
    set using set_precision.  Note: returned geometry does not have precision
    set unless specified previously by set_precision.

    Parameters
    ----------
    a : Geometry or array_like
        Geometry or geometries to subtract b from.
    b : Geometry or array_like
        Geometry or geometries to subtract from a.
    grid_size : float, optional
        Precision grid size; will use the highest precision of the inputs by default.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``grid_size`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    set_precision

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line = LineString([(0, 0), (2, 2)])
    >>> shapely.difference(line, LineString([(1, 1), (3, 3)]))
    <LINESTRING (0 0, 1 1)>
    >>> shapely.difference(line, LineString())
    <LINESTRING (0 0, 2 2)>
    >>> shapely.difference(line, None) is None
    True
    >>> box1 = shapely.box(0, 0, 2, 2)
    >>> box2 = shapely.box(1, 1, 3, 3)
    >>> shapely.difference(box1, box2).normalize()
    <POLYGON ((0 0, 0 2, 1 2, 1 1, 2 1, 2 0, 0 0))>
    >>> box1 = shapely.box(0.1, 0.2, 2.1, 2.1)
    >>> shapely.difference(box1, box2, grid_size=1)
    <POLYGON ((2 0, 0 0, 0 2, 1 2, 1 1, 2 1, 2 0))>

    """
    if grid_size is not None:
        if not np.isscalar(grid_size):
            raise ValueError("grid_size parameter only accepts scalar values")

        return lib.difference_prec(a, b, grid_size, **kwargs)

    return lib.difference(a, b, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   intersection(a, b, grid_size=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'grid_size' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'b'
#   intersection(a, b, *, grid_size=None, **kwargs)


@deprecate_positional(["grid_size"], category=DeprecationWarning)
@multithreading_enabled
def intersection(a, b, grid_size=None, **kwargs):
    """Return the geometry that is shared between input geometries.

    If grid_size is nonzero, input coordinates will be snapped to a precision
    grid of that size and resulting coordinates will be snapped to that same
    grid.  If 0, this operation will use double precision coordinates.  If None,
    the highest precision of the inputs will be used, which may be previously
    set using set_precision.  Note: returned geometry does not have precision
    set unless specified previously by set_precision.

    Parameters
    ----------
    a, b : Geometry or array_like
        Geometry or geometries to intersect with.
    grid_size : float, optional
        Precision grid size; will use the highest precision of the inputs by default.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``grid_size`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    intersection_all
    set_precision

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line = LineString([(0, 0), (2, 2)])
    >>> shapely.intersection(line, LineString([(1, 1), (3, 3)]))
    <LINESTRING (1 1, 2 2)>
    >>> box1 = shapely.box(0, 0, 2, 2)
    >>> box2 = shapely.box(1, 1, 3, 3)
    >>> shapely.intersection(box1, box2).normalize()
    <POLYGON ((1 1, 1 2, 2 2, 2 1, 1 1))>
    >>> box1 = shapely.box(0.1, 0.2, 2.1, 2.1)
    >>> shapely.intersection(box1, box2, grid_size=1)
    <POLYGON ((2 2, 2 1, 1 1, 1 2, 2 2))>

    """
    if grid_size is not None:
        if not np.isscalar(grid_size):
            raise ValueError("grid_size parameter only accepts scalar values")

        return lib.intersection_prec(a, b, grid_size, **kwargs)

    return lib.intersection(a, b, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   intersection_all(geometries, axis=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'axis' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'geometries'
#   intersection_all(geometries, *, axis=None, **kwargs)


@deprecate_positional(["axis"], category=DeprecationWarning)
@multithreading_enabled
def intersection_all(geometries, axis=None, **kwargs):
    """Return the intersection of multiple geometries.

    This function ignores None values when other Geometry elements are present.
    If all elements of the given axis are None, an empty GeometryCollection is
    returned.

    Parameters
    ----------
    geometries : array_like
        Geometries to calculate the intersection of.
    axis : int, optional
        Axis along which the operation is performed. The default (None)
        performs the operation over all axes, returning a scalar value.
        Axis may be negative, in which case it counts from the last to the
        first axis.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``axis`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    intersection

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line1 = LineString([(0, 0), (2, 2)])
    >>> line2 = LineString([(1, 1), (3, 3)])
    >>> shapely.intersection_all([line1, line2])
    <LINESTRING (1 1, 2 2)>
    >>> shapely.intersection_all([[line1, line2, None]], axis=1).tolist()
    [<LINESTRING (1 1, 2 2)>]
    >>> shapely.intersection_all([line1, None])
    <LINESTRING (0 0, 2 2)>

    """
    geometries = np.asarray(geometries)
    if axis is None:
        geometries = geometries.ravel()
    else:
        geometries = np.rollaxis(geometries, axis=axis, start=geometries.ndim)

    return lib.intersection_all(geometries, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   symmetric_difference(a, b, grid_size=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'grid_size' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'b'
#   symmetric_difference(a, b, *, grid_size=None, **kwargs)


@deprecate_positional(["grid_size"], category=DeprecationWarning)
@multithreading_enabled
def symmetric_difference(a, b, grid_size=None, **kwargs):
    """Return the geometry with the portions of input geometries that do not intersect.

    If grid_size is nonzero, input coordinates will be snapped to a precision
    grid of that size and resulting coordinates will be snapped to that same
    grid.  If 0, this operation will use double precision coordinates.  If None,
    the highest precision of the inputs will be used, which may be previously
    set using set_precision.  Note: returned geometry does not have precision
    set unless specified previously by set_precision.

    Parameters
    ----------
    a, b : Geometry or array_like
        Geometry or geometries to evaluate symmetric difference with.
    grid_size : float, optional
        Precision grid size; will use the highest precision of the inputs by default.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``grid_size`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    symmetric_difference_all
    set_precision

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line = LineString([(0, 0), (2, 2)])
    >>> shapely.symmetric_difference(line, LineString([(1, 1), (3, 3)]))
    <MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))>
    >>> box1 = shapely.box(0, 0, 2, 2)
    >>> box2 = shapely.box(1, 1, 3, 3)
    >>> shapely.symmetric_difference(box1, box2).normalize()
    <MULTIPOLYGON (((1 2, 1 3, 3 3, 3 1, 2 1, 2 2, 1 2)), ((0 0, 0 2, 1 2, 1 1, ...>
    >>> box1 = shapely.box(0.1, 0.2, 2.1, 2.1)
    >>> shapely.symmetric_difference(box1, box2, grid_size=1)
    <MULTIPOLYGON (((2 0, 0 0, 0 2, 1 2, 1 1, 2 1, 2 0)), ((2 2, 1 2, 1 3, 3 3, ...>

    """
    if grid_size is not None:
        if not np.isscalar(grid_size):
            raise ValueError("grid_size parameter only accepts scalar values")

        return lib.symmetric_difference_prec(a, b, grid_size, **kwargs)

    return lib.symmetric_difference(a, b, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   symmetric_difference_all(geometries, axis=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'axis' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'geometries'
#   symmetric_difference_all(geometries, *, axis=None, **kwargs)


@deprecate_positional(["axis"], category=DeprecationWarning)
@multithreading_enabled
def symmetric_difference_all(geometries, axis=None, **kwargs):
    """Return the symmetric difference of multiple geometries.

    This function ignores None values when other Geometry elements are present.
    If all elements of the given axis are None an empty GeometryCollection is
    returned.

    .. deprecated:: 2.1.0

        This function behaves incorrectly and will be removed in a future
        version. See https://github.com/shapely/shapely/issues/2027 for more
        details.

    Parameters
    ----------
    geometries : array_like
        Geometries to calculate the combined symmetric difference of.
    axis : int, optional
        Axis along which the operation is performed. The default (None)
        performs the operation over all axes, returning a scalar value.
        Axis may be negative, in which case it counts from the last to the
        first axis.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``axis`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    symmetric_difference

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line1 = LineString([(0, 0), (2, 2)])
    >>> line2 = LineString([(1, 1), (3, 3)])
    >>> shapely.symmetric_difference_all([line1, line2])
    <MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))>
    >>> shapely.symmetric_difference_all([[line1, line2, None]], axis=1).tolist()
    [<MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))>]
    >>> shapely.symmetric_difference_all([line1, None])
    <LINESTRING (0 0, 2 2)>
    >>> shapely.symmetric_difference_all([None, None])
    <GEOMETRYCOLLECTION EMPTY>

    """
    warnings.warn(
        "The symmetric_difference_all function behaves incorrectly and will be "
        "removed in a future version. "
        "See https://github.com/shapely/shapely/issues/2027 for more details.",
        DeprecationWarning,
        stacklevel=2,
    )
    geometries = np.asarray(geometries)
    if axis is None:
        geometries = geometries.ravel()
    else:
        geometries = np.rollaxis(geometries, axis=axis, start=geometries.ndim)

    return lib.symmetric_difference_all(geometries, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   union(a, b, grid_size=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'grid_size' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'b'
#   union(a, b, *, grid_size=None, **kwargs)


@deprecate_positional(["grid_size"], category=DeprecationWarning)
@multithreading_enabled
def union(a, b, grid_size=None, **kwargs):
    """Merge geometries into one.

    If grid_size is nonzero, input coordinates will be snapped to a precision
    grid of that size and resulting coordinates will be snapped to that same
    grid.  If 0, this operation will use double precision coordinates.  If None,
    the highest precision of the inputs will be used, which may be previously
    set using set_precision.  Note: returned geometry does not have precision
    set unless specified previously by set_precision.

    Parameters
    ----------
    a, b : Geometry or array_like
        Geometry or geometries to merge (union).
    grid_size : float, optional
        Precision grid size; will use the highest precision of the inputs by default.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``grid_size`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    union_all
    set_precision

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString
    >>> line = LineString([(0, 0), (2, 2)])
    >>> shapely.union(line, LineString([(2, 2), (3, 3)]))
    <MULTILINESTRING ((0 0, 2 2), (2 2, 3 3))>
    >>> shapely.union(line, None) is None
    True
    >>> box1 = shapely.box(0, 0, 2, 2)
    >>> box2 = shapely.box(1, 1, 3, 3)
    >>> shapely.union(box1, box2).normalize()
    <POLYGON ((0 0, 0 2, 1 2, 1 3, 3 3, 3 1, 2 1, 2 0, 0 0))>
    >>> box1 = shapely.box(0.1, 0.2, 2.1, 2.1)
    >>> shapely.union(box1, box2, grid_size=1)
    <POLYGON ((2 0, 0 0, 0 2, 1 2, 1 3, 3 3, 3 1, 2 1, 2 0))>

    """
    if grid_size is not None:
        if not np.isscalar(grid_size):
            raise ValueError("grid_size parameter only accepts scalar values")

        return lib.union_prec(a, b, grid_size, **kwargs)

    return lib.union(a, b, **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   union_all(geometries, grid_size=None, axis=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'grid_size' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'geometries'
#   union_all(geometries, *, grid_size=None, axis=None, **kwargs)


@deprecate_positional(["grid_size", "axis"], category=DeprecationWarning)
@multithreading_enabled
def union_all(geometries, grid_size=None, axis=None, **kwargs):
    """Return the union of multiple geometries.

    This function ignores None values when other Geometry elements are present.
    If all elements of the given axis are None an empty GeometryCollection is
    returned.

    If grid_size is nonzero, input coordinates will be snapped to a precision
    grid of that size and resulting coordinates will be snapped to that same
    grid.  If 0, this operation will use double precision coordinates.  If None,
    the highest precision of the inputs will be used, which may be previously
    set using set_precision.  Note: returned geometry does not have precision
    set unless specified previously by set_precision.

    `unary_union` is an alias of `union_all`.

    Parameters
    ----------
    geometries : array_like
        Geometries to merge/union.
    grid_size : float, optional
        Precision grid size; will use the highest precision of the inputs by default.
    axis : int, optional
        Axis along which the operation is performed. The default (None)
        performs the operation over all axes, returning a scalar value.
        Axis may be negative, in which case it counts from the last to the
        first axis.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``grid_size`` or ``axis`` are
        specified as positional arguments. In a future release, these will
        need to be specified as keyword arguments.

    See Also
    --------
    union
    set_precision

    Examples
    --------
    >>> import shapely
    >>> from shapely import LineString, Point
    >>> line1 = LineString([(0, 0), (2, 2)])
    >>> line2 = LineString([(2, 2), (3, 3)])
    >>> shapely.union_all([line1, line2])
    <MULTILINESTRING ((0 0, 2 2), (2 2, 3 3))>
    >>> shapely.union_all([[line1, line2, None]], axis=1).tolist()
    [<MULTILINESTRING ((0 0, 2 2), (2 2, 3 3))>]
    >>> box1 = shapely.box(0, 0, 2, 2)
    >>> box2 = shapely.box(1, 1, 3, 3)
    >>> shapely.union_all([box1, box2]).normalize()
    <POLYGON ((0 0, 0 2, 1 2, 1 3, 3 3, 3 1, 2 1, 2 0, 0 0))>
    >>> box1 = shapely.box(0.1, 0.2, 2.1, 2.1)
    >>> shapely.union_all([box1, box2], grid_size=1)
    <POLYGON ((2 0, 0 0, 0 2, 1 2, 1 3, 3 3, 3 1, 2 1, 2 0))>
    >>> shapely.union_all([None, Point(0, 1)])
    <POINT (0 1)>
    >>> shapely.union_all([None, None])
    <GEOMETRYCOLLECTION EMPTY>
    >>> shapely.union_all([])
    <GEOMETRYCOLLECTION EMPTY>

    """
    # for union_all, GEOS provides an efficient route through first creating
    # GeometryCollections
    # first roll the aggregation axis backwards
    geometries = np.asarray(geometries)
    if axis is None:
        geometries = geometries.ravel()
    else:
        geometries = np.rollaxis(geometries, axis=axis, start=geometries.ndim)

    # create_collection acts on the inner axis
    collections = lib.create_collection(
        geometries, np.intc(GeometryType.GEOMETRYCOLLECTION)
    )

    if grid_size is not None:
        if not np.isscalar(grid_size):
            raise ValueError("grid_size parameter only accepts scalar values")

        return lib.unary_union_prec(collections, grid_size, **kwargs)

    return lib.unary_union(collections, **kwargs)


unary_union = union_all


@multithreading_enabled
def coverage_union(a, b, **kwargs):
    """Merge multiple polygons into one.

    This is an optimized version of union which assumes the polygons to be
    non-overlapping.
    If this assumption is not met, the exact result is not guaranteed
    (depending on the GEOS version, it may return the input unchanged or raise
    an error).

    Parameters
    ----------
    a, b : Geometry or array_like
        Geometry or geometries to merge (union).
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    See Also
    --------
    coverage_union_all

    Examples
    --------
    >>> import shapely
    >>> from shapely import Polygon
    >>> polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    >>> polygon_2 = Polygon([(1, 0), (1, 1), (2, 1), (2, 0), (1, 0)])
    >>> shapely.coverage_union(polygon_1, polygon_2).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 2 1, 2 0, 1 0, 0 0))>

    Union with None returns same polygon

    >>> shapely.coverage_union(polygon_1, None).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))>

    """
    return coverage_union_all([a, b], **kwargs)


# Note: future plan is to change this signature over a few releases:
# shapely 2.0:
#   coverage_union_all(geometries, axis=None, **kwargs)
# shapely 2.1: shows deprecation warning about positional 'axis' arg
#   same signature as 2.0
# shapely 2.2(?): enforce keyword-only arguments after 'geometries'
#   coverage_union_all(geometries, *, axis=None, **kwargs)


@deprecate_positional(["axis"], category=DeprecationWarning)
@multithreading_enabled
def coverage_union_all(geometries, axis=None, **kwargs):
    """Return the union of multiple polygons of a geometry collection.

    This is an optimized version of union which assumes the polygons
    to be non-overlapping.

    This function ignores None values when other Geometry elements are present.
    If all elements of the given axis are None, an empty GeometryCollection is
    returned (before GEOS 3.12 this was an empty MultiPolygon).

    Parameters
    ----------
    geometries : array_like
        Geometries to merge/union.
    axis : int, optional
        Axis along which the operation is performed. The default (None)
        performs the operation over all axes, returning a scalar value.
        Axis may be negative, in which case it counts from the last to the
        first axis.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    Notes
    -----

    .. deprecated:: 2.1.0
        A deprecation warning is shown if ``axis`` is specified as a
        positional argument. This will need to be specified as a keyword
        argument in a future release.

    See Also
    --------
    coverage_union

    Examples
    --------
    >>> import shapely
    >>> from shapely import Polygon
    >>> polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    >>> polygon_2 = Polygon([(1, 0), (1, 1), (2, 1), (2, 0), (1, 0)])
    >>> shapely.coverage_union_all([polygon_1, polygon_2]).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 2 1, 2 0, 1 0, 0 0))>
    >>> shapely.coverage_union_all([polygon_1, None]).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))>
    >>> shapely.coverage_union_all([None, None]).normalize()
    <GEOMETRYCOLLECTION EMPTY>

    """
    # coverage union in GEOS works over GeometryCollections
    # first roll the aggregation axis backwards
    geometries = np.asarray(geometries)
    if axis is None:
        geometries = geometries.ravel()
    else:
        geometries = np.rollaxis(
            np.asarray(geometries), axis=axis, start=geometries.ndim
        )
    # create_collection acts on the inner axis
    collections = lib.create_collection(
        geometries, np.intc(GeometryType.GEOMETRYCOLLECTION)
    )
    return lib.coverage_union(collections, **kwargs)


@requires_geos("3.12.0")
@multithreading_enabled
def disjoint_subset_union(a, b, **kwargs):
    """Merge multiple polygons into one using algorithm optimised for subsets.

    This is an optimized version of union which assumes inputs can be
    divided into subsets that do not intersect.

    If there is only one such subset, performance can be expected to be worse than
    :func:`union`. As such, it is recommeded to use ``disjoint_subset_union`` with
    GeometryCollections rather than individual geometries.

    .. versionadded:: 2.1.0

    Parameters
    ----------
    a, b : Geometry or array_like
        Geometry or geometries to merge (union).
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    See Also
    --------
    union
    coverage_union
    disjoint_subset_union_all

    Examples
    --------
    >>> import shapely
    >>> from shapely import Polygon
    >>> polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    >>> polygon_2 = Polygon([(1, 0), (1, 1), (2, 1), (2, 0), (1, 0)])
    >>> shapely.disjoint_subset_union(polygon_1, polygon_2).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 2 1, 2 0, 1 0, 0 0))>

    Union with None returns same polygon:

    >>> shapely.disjoint_subset_union(polygon_1, None).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))>
    """
    if (isinstance(a, Geometry) or a is None) and (
        isinstance(b, Geometry) or b is None
    ):
        pass
    elif isinstance(a, Geometry) or a is None:
        a = np.full_like(b, a)
    elif isinstance(b, Geometry) or b is None:
        b = np.full_like(a, b)
    elif len(a) != len(b):
        raise ValueError("Arrays a and b must have the same length")
    return disjoint_subset_union_all([a, b], axis=0, **kwargs)


@requires_geos("3.12.0")
@multithreading_enabled
def disjoint_subset_union_all(geometries, *, axis=None, **kwargs):
    """Return the union of multiple polygons.

    This is an optimized version of union which assumes inputs can be divided into
    subsets that do not intersect.

    If there is only one such subset, performance can be expected to be worse than
    :func:`union_all`.

    This function ignores None values when other Geometry elements are present.
    If all elements of the given axis are None, an empty GeometryCollection is
    returned.

    .. versionadded:: 2.1.0

    Parameters
    ----------
    geometries : array_like
        Geometries to union.
    axis : int, optional
        Axis along which the operation is performed. The default (None)
        performs the operation over all axes, returning a scalar value.
        Axis may be negative, in which case it counts from the last to the
        first axis.
    **kwargs
        See :ref:`NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    See Also
    --------
    coverage_union_all
    union_all
    disjoint_subset_union

    Examples
    --------
    >>> import shapely
    >>> from shapely import Polygon
    >>> polygon_1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    >>> polygon_2 = Polygon([(1, 0), (1, 1), (2, 1), (2, 0), (1, 0)])
    >>> shapely.disjoint_subset_union_all([polygon_1, polygon_2]).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 2 1, 2 0, 1 0, 0 0))>
    >>> shapely.disjoint_subset_union_all([polygon_1, None]).normalize()
    <POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))>
    >>> shapely.disjoint_subset_union_all([None, None]).normalize()
    <GEOMETRYCOLLECTION EMPTY>
    """
    geometries = np.asarray(geometries)
    if axis is None:
        geometries = geometries.ravel()
    else:
        geometries = np.rollaxis(
            np.asarray(geometries), axis=axis, start=geometries.ndim
        )
    # create_collection acts on the inner axis
    collections = lib.create_collection(
        geometries, np.intc(GeometryType.GEOMETRYCOLLECTION)
    )

    return lib.disjoint_subset_union(collections, **kwargs)



import numpy as np
import pytest

import shapely
from shapely import LinearRing, LineString, MultiLineString, Point, Polygon
from shapely.tests.common import all_types, all_types_z, empty_point, ignore_invalid

all_non_empty_types = np.array(all_types + all_types_z)[
    ~shapely.is_empty(all_types + all_types_z)
]


# TODO add all_types_m and all_types_zm once tranform supports M coordinates
@pytest.mark.parametrize("geom", all_types + all_types_z)
def test_equality(geom):
    assert geom == geom  # noqa: PLR0124
    transformed = shapely.transform(geom, lambda x: x, include_z=True)
    if (
        shapely.geos_version < (3, 9, 0)
        and isinstance(geom, Point)
        and geom.is_empty
        and not geom.has_z
    ):
        # the transformed empty 2D point has become 3D on GEOS 3.8
        transformed = shapely.force_2d(geom)
    assert geom == transformed
    assert not (geom != transformed)


@pytest.mark.parametrize(
    "left, right",
    # automated test cases with transformed coordinate values
    [(geom, shapely.transform(geom, lambda x: x + 1)) for geom in all_non_empty_types]
    + [
        # (slightly) different coordinate values
        (LineString([(0, 0), (1, 1)]), LineString([(0, 0), (1, 2)])),
        (LineString([(0, 0), (1, 1)]), LineString([(0, 0), (1, 1 + 1e-12)])),
        # different coordinate order
        (LineString([(0, 0), (1, 1)]), LineString([(1, 1), (0, 0)])),
        # different number of coordinates (but spatially equal)
        (LineString([(0, 0), (1, 1)]), LineString([(0, 0), (1, 1), (1, 1)])),
        (LineString([(0, 0), (1, 1)]), LineString([(0, 0), (0.5, 0.5), (1, 1)])),
        # different order of sub-geometries
        (
            MultiLineString([[(1, 1), (2, 2)], [(2, 2), (3, 3)]]),
            MultiLineString([[(2, 2), (3, 3)], [(1, 1), (2, 2)]]),
        ),
        # M coordinates (don't work yet with automated cases)
        pytest.param(
            shapely.from_wkt("POINT M (0 0 0)"),
            shapely.from_wkt("POINT M (0 0 1)"),
            marks=pytest.mark.skipif(
                shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12"
            ),
        ),
        pytest.param(
            shapely.from_wkt("POINT ZM (0 0 0 0)"),
            shapely.from_wkt("POINT ZM (0 0 0 1)"),
            marks=pytest.mark.skipif(
                shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12"
            ),
        ),
    ],
)
def test_equality_false(left, right):
    assert left != right
    assert not (left == right)


with ignore_invalid():
    cases1 = [
        (LineString([(0, 1), (2, np.nan)]), LineString([(0, 1), (2, np.nan)])),
        (
            LineString([(0, 1), (np.nan, np.nan)]),
            LineString([(0, 1), (np.nan, np.nan)]),
        ),
        (LineString([(np.nan, 1), (2, 3)]), LineString([(np.nan, 1), (2, 3)])),
        (LineString([(0, np.nan), (2, 3)]), LineString([(0, np.nan), (2, 3)])),
        (
            LineString([(np.nan, np.nan), (np.nan, np.nan)]),
            LineString([(np.nan, np.nan), (np.nan, np.nan)]),
        ),
        # NaN as explicit Z coordinate
        # TODO: if first z is NaN -> considered as 2D -> tested below explicitly
        # (
        #     LineString([(0, 1, np.nan), (2, 3, np.nan)]),
        #     LineString([(0, 1, np.nan), (2, 3, np.nan)]),
        # ),
        (
            LineString([(0, 1, 2), (2, 3, np.nan)]),
            LineString([(0, 1, 2), (2, 3, np.nan)]),
        ),
        # (
        #     LineString([(0, 1, np.nan), (2, 3, 4)]),
        #     LineString([(0, 1, np.nan), (2, 3, 4)]),
        # ),
    ]


@pytest.mark.parametrize("left, right", cases1)
def test_equality_with_nan(left, right):
    assert left == right
    assert not (left != right)


with ignore_invalid():
    cases2 = [
        (
            LineString([(0, 1, np.nan), (2, 3, np.nan)]),
            LineString([(0, 1, np.nan), (2, 3, np.nan)]),
        ),
        (
            LineString([(0, 1, np.nan), (2, 3, 4)]),
            LineString([(0, 1, np.nan), (2, 3, 4)]),
        ),
    ]


@pytest.mark.parametrize("left, right", cases2)
def test_equality_with_nan_z(left, right):
    assert left == right
    assert not (left != right)


with ignore_invalid():
    cases3 = [
        (LineString([(0, np.nan), (2, 3)]), LineString([(0, 1), (2, 3)])),
        (LineString([(0, 1), (2, np.nan)]), LineString([(0, 1), (2, 3)])),
        (LineString([(0, 1, np.nan), (2, 3, 4)]), LineString([(0, 1, 2), (2, 3, 4)])),
        (LineString([(0, 1, 2), (2, 3, np.nan)]), LineString([(0, 1, 2), (2, 3, 4)])),
        pytest.param(
            shapely.from_wkt("POINT M (0 0 0)"),
            shapely.from_wkt("POINT M (0 0 NaN)"),
            marks=pytest.mark.skipif(
                shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12"
            ),
        ),
        pytest.param(
            shapely.from_wkt("POINT ZM (0 0 0 0)"),
            shapely.from_wkt("POINT ZM (0 0 0 NaN)"),
            marks=pytest.mark.skipif(
                shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12"
            ),
        ),
    ]


@pytest.mark.parametrize("left, right", cases3)
def test_equality_with_nan_false(left, right):
    assert left != right


def test_equality_with_nan_z_false():
    with ignore_invalid():
        left = LineString([(0, 1, np.nan), (2, 3, np.nan)])
        right = LineString([(0, 1, np.nan), (2, 3, 4)])

    if shapely.geos_version < (3, 10, 0):
        # GEOS <= 3.9 fill the NaN with 0, so the z dimension is different
        assert left != right
    elif shapely.geos_version < (3, 12, 0):
        # GEOS 3.10-3.11 ignore NaN for Z also when explicitly created with 3D
        # and so the geometries are considered as 2D (and thus z dimension is ignored)
        assert left == right
    else:
        assert left != right


def test_equality_z():
    # different dimensionality
    geom1 = Point(0, 1)
    geom2 = Point(0, 1, 0)
    assert geom1 != geom2

    # different dimensionality with NaN z
    geom2 = Point(0, 1, np.nan)
    if shapely.geos_version < (3, 12, 0):
        # GEOS 3.10-3.11 ignore NaN for Z also when explicitly created with 3D
        # and so the geometries are considered as 2D (and thus z dimension is ignored)
        assert geom1 == geom2
    else:
        assert geom1 != geom2


def test_equality_exact_type():
    # geometries with different type but same coord seq are not equal
    geom1 = LineString([(0, 0), (1, 1), (0, 1), (0, 0)])
    geom2 = LinearRing([(0, 0), (1, 1), (0, 1), (0, 0)])
    geom3 = Polygon([(0, 0), (1, 1), (0, 1), (0, 0)])
    assert geom1 != geom2
    assert geom1 != geom3
    assert geom2 != geom3

    # empty with different type
    geom1 = shapely.from_wkt("POINT EMPTY")
    geom2 = shapely.from_wkt("LINESTRING EMPTY")
    assert geom1 != geom2


def test_equality_polygon():
    # different exterior rings
    geom1 = shapely.from_wkt("POLYGON ((0 0, 10 0, 10 10, 0 10, 0 0))")
    geom2 = shapely.from_wkt("POLYGON ((0 0, 10 0, 10 10, 0 15, 0 0))")
    assert geom1 != geom2

    # different number of holes
    geom1 = shapely.from_wkt(
        "POLYGON ((0 0, 10 0, 10 10, 0 10, 0 0), (1 1, 2 1, 2 2, 1 1))"
    )
    geom2 = shapely.from_wkt(
        "POLYGON ((0 0, 10 0, 10 10, 0 10, 0 0), (1 1, 2 1, 2 2, 1 1), "
        "(3 3, 4 3, 4 4, 3 3))"
    )
    assert geom1 != geom2

    # different order of holes
    geom1 = shapely.from_wkt(
        "POLYGON ((0 0, 10 0, 10 10, 0 10, 0 0), (3 3, 4 3, 4 4, 3 3), "
        "(1 1, 2 1, 2 2, 1 1))"
    )
    geom2 = shapely.from_wkt(
        "POLYGON ((0 0, 10 0, 10 10, 0 10, 0 0), (1 1, 2 1, 2 2, 1 1), "
        "(3 3, 4 3, 4 4, 3 3))"
    )
    assert geom1 != geom2


@pytest.mark.parametrize("geom", all_types)
def test_comparison_notimplemented(geom):
    # comparing to a non-geometry class should return NotImplemented in __eq__
    # to ensure proper delegation to other (eg to ensure comparison of scalar
    # with array works)
    # https://github.com/shapely/shapely/issues/1056
    assert geom.__eq__(1) is NotImplemented

    # with array
    arr = np.array([geom, geom], dtype=object)

    result = arr == geom
    assert isinstance(result, np.ndarray)
    assert result.all()

    result = geom == arr
    assert isinstance(result, np.ndarray)
    assert result.all()

    result = arr != geom
    assert isinstance(result, np.ndarray)
    assert not result.any()

    result = geom != arr
    assert isinstance(result, np.ndarray)
    assert not result.any()


def test_comparison_not_supported():
    geom1 = Point(1, 1)
    geom2 = Point(2, 2)

    with pytest.raises(TypeError, match="not supported between instances"):
        assert geom1 > geom2

    with pytest.raises(TypeError, match="not supported between instances"):
        assert geom1 < geom2

    with pytest.raises(TypeError, match="not supported between instances"):
        assert geom1 >= geom2

    with pytest.raises(TypeError, match="not supported between instances"):
        assert geom1 <= geom2


@pytest.mark.parametrize(
    "geom", all_types + (shapely.points(np.nan, np.nan), empty_point)
)
def test_hash_same_equal(geom):
    hash1 = hash(geom)
    hash2 = hash(shapely.transform(geom, lambda x: x))
    assert hash1 == hash2, geom


@pytest.mark.parametrize("geom", all_non_empty_types)
def test_hash_same_not_equal(geom):
    assert hash(geom) != hash(shapely.transform(geom, lambda x: x + 1))



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

import shapely
from shapely import count_coordinates, get_coordinates, set_coordinates, transform
from shapely.tests.common import (
    empty,
    empty_line_string_z,
    empty_point,
    empty_point_m,
    empty_point_z,
    empty_point_zm,
    geometry_collection,
    geometry_collection_z,
    geometry_collection_zm,
    line_string,
    line_string_m,
    line_string_z,
    line_string_zm,
    linear_ring,
    multi_line_string,
    multi_point,
    multi_polygon,
    point,
    point_m,
    point_z,
    point_zm,
    polygon,
    polygon_with_hole,
    polygon_z,
)

nested_2 = shapely.geometrycollections([geometry_collection, point])
nested_3 = shapely.geometrycollections([nested_2, point])


@pytest.mark.parametrize(
    "geoms,count",
    [
        ([], 0),
        ([empty], 0),
        ([point, empty], 1),
        ([empty, point, empty], 1),
        ([point, None], 1),
        ([None, point, None], 1),
        ([point, point], 2),
        ([point, point_z], 2),
        ([line_string, linear_ring], 8),
        ([polygon], 5),
        ([polygon_with_hole], 10),
        ([multi_point, multi_line_string], 4),
        ([multi_polygon], 10),
        ([geometry_collection], 3),
        ([nested_2], 4),
        ([nested_3], 5),
    ],
)
def test_count_coords(geoms, count):
    actual = count_coordinates(np.array(geoms, np.object_))
    assert actual == count


@pytest.mark.parametrize("include_m", [True, False])
@pytest.mark.parametrize("include_z", [True, False])
@pytest.mark.parametrize(
    "geoms,x,y",
    [
        ([], [], []),
        ([empty], [], []),
        ([point, empty], [2], [3]),
        ([empty, point, empty], [2], [3]),
        ([point, None], [2], [3]),
        ([None, point, None], [2], [3]),
        ([point, point], [2, 2], [3, 3]),
        (
            [line_string, linear_ring],
            [0, 1, 1, 0, 1, 1, 0, 0],
            [0, 0, 1, 0, 0, 1, 1, 0],
        ),
        ([polygon], [0, 2, 2, 0, 0], [0, 0, 2, 2, 0]),
        (
            [polygon_with_hole],
            [0, 0, 10, 10, 0, 2, 2, 4, 4, 2],
            [0, 10, 10, 0, 0, 2, 4, 4, 2, 2],
        ),
        ([multi_point, multi_line_string], [0, 1, 0, 1], [0, 2, 0, 2]),
        (
            [multi_polygon],
            [0, 1, 1, 0, 0, 2.1, 2.2, 2.2, 2.1, 2.1],
            [0, 0, 1, 1, 0, 2.1, 2.1, 2.2, 2.2, 2.1],
        ),
        ([geometry_collection], [51, 52, 49], [-1, -1, 2]),
        ([nested_2], [51, 52, 49, 2], [-1, -1, 2, 3]),
        ([nested_3], [51, 52, 49, 2, 2], [-1, -1, 2, 3, 3]),
    ],
)
def test_get_coords(geoms, x, y, include_z, include_m):
    actual = get_coordinates(geoms, include_z=include_z, include_m=include_m)
    expected = [x, y]
    if include_z:
        expected.append([np.nan] * len(x))
    if include_m:
        expected.append([np.nan] * len(x))
    assert_equal(actual, np.array(expected, np.float64).T)


@pytest.mark.parametrize(
    "geoms,index",
    [
        ([], []),
        ([empty], []),
        ([point, empty], [0]),
        ([empty, point, empty], [1]),
        ([point, None], [0]),
        ([None, point, None], [1]),
        ([point, point], [0, 1]),
        ([point, line_string], [0, 1, 1, 1]),
        ([line_string, point], [0, 0, 0, 1]),
        ([line_string, linear_ring], [0, 0, 0, 1, 1, 1, 1, 1]),
    ],
)
def test_get_coords_index(geoms, index):
    _, actual = get_coordinates(np.array(geoms, np.object_), return_index=True)
    expected = np.array(index, dtype=np.intp)
    assert_equal(actual, expected)


@pytest.mark.parametrize("order", ["C", "F"])
def test_get_coords_index_multidim(order):
    geometry = np.array([[point, line_string], [empty, empty]], order=order)
    expected = [0, 1, 1, 1]  # would be [0, 2, 2, 2] with fortran order
    _, actual = get_coordinates(geometry, return_index=True)
    assert_equal(actual, expected)


@pytest.mark.parametrize("include_m", [True, False])
@pytest.mark.parametrize("include_z", [True, False])
@pytest.mark.parametrize(
    "geoms,x,y,z",
    [
        ([point, point_z], [2, 2], [3, 3], [np.nan, 4]),
        ([line_string_z], [0, 1, 1], [0, 0, 1], [4, 4, 4]),
        ([polygon_z], [0, 2, 2, 0, 0], [0, 0, 2, 2, 0], [4, 4, 4, 4, 4]),
        ([geometry_collection_z], [2, 0, 1, 1], [3, 0, 0, 1], [4, 4, 4, 4]),
        ([point, empty_point], [2], [3], [np.nan]),
    ],
)
def test_get_coords_z(geoms, x, y, z, include_z, include_m):
    actual = get_coordinates(geoms, include_z=include_z, include_m=include_m)
    expected = [x, y]
    if include_z:
        expected.append(z)
    if include_m:
        expected.append([np.nan] * len(x))
    assert_equal(actual, np.array(expected, np.float64).T)


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
@pytest.mark.parametrize("include_m", [True, False])
@pytest.mark.parametrize("include_z", [True, False])
@pytest.mark.parametrize(
    "geoms,x,y,z,m",
    [
        (
            [point, point_z, point_m, point_zm],
            [2, 2, 2, 2],
            [3, 3, 3, 3],
            [np.nan, 4, np.nan, 4],
            [np.nan, np.nan, 5, 5],
        ),
        (
            [line_string, line_string_z, line_string_m, line_string_zm],
            [0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1],
            [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
            [np.nan, np.nan, np.nan, 4, 4, 4] * 2,
            [np.nan] * 6 + [1, 2, 3, 1, 2, 3],
        ),
        (
            [geometry_collection_zm],
            [2, 0, 1, 1],
            [3, 0, 0, 1],
            [4, 4, 4, 4],
            [5, 1, 2, 3],
        ),
        (
            [point, empty_point, empty_point_z, empty_point_m, empty_point_zm],
            [2],
            [3],
            [np.nan],
            [np.nan],
        ),
    ],
)
def test_get_coords_zm(geoms, x, y, z, m, include_z, include_m):
    actual = get_coordinates(geoms, include_z=include_z, include_m=include_m)
    expected = [x, y]
    if include_z:
        expected.append(z)
    if include_m:
        expected.append(m)
    assert_equal(actual, np.array(expected, np.float64).T)


def test_get_coords_deprecate_positional():
    with pytest.deprecated_call(
        match="positional argument `include_z` for `get_coordinates` is deprecated"
    ):
        get_coordinates(point, False)
    with pytest.deprecated_call(
        match="positional arguments `include_z` and `return_index` "
        "for `get_coordinates` are deprecated"
    ):
        get_coordinates(point, False, False)


@pytest.mark.parametrize("include_z", [True, False])
@pytest.mark.parametrize(
    "geoms,count,has_ring",
    [
        ([], 0, False),
        ([empty], 0, False),
        ([empty_point], 0, False),
        ([point, empty], 1, False),
        ([empty, point, empty], 1, False),
        ([point, None], 1, False),
        ([None, point, None], 1, False),
        ([point, point], 2, False),
        ([point, point_z], 2, False),
        ([line_string, linear_ring], 8, True),
        ([line_string_z], 3, True),
        ([polygon], 5, True),
        ([polygon_z], 5, True),
        ([polygon_with_hole], 10, True),
        ([multi_point, multi_line_string], 4, False),
        ([multi_polygon], 10, True),
        ([geometry_collection], 3, False),
        ([geometry_collection_z], 3, False),
        ([nested_2], 4, False),
        ([nested_3], 5, False),
    ],
)
def test_set_coords(geoms, count, has_ring, include_z):
    arr_geoms = np.array(geoms, np.object_)
    n = 3 if include_z else 2
    coords = get_coordinates(arr_geoms, include_z=include_z) + np.random.random((1, n))
    new_geoms = set_coordinates(arr_geoms, coords)
    assert_equal(coords, get_coordinates(new_geoms, include_z=include_z))


def test_set_coords_nan():
    geoms = np.array([point])
    coords = np.array([[np.nan, np.inf]])
    new_geoms = set_coordinates(geoms, coords)
    assert_equal(coords, get_coordinates(new_geoms))


def test_set_coords_breaks_ring():
    with pytest.raises(shapely.GEOSException):
        set_coordinates(linear_ring, np.random.random((5, 2)))


def test_set_coords_0dim():
    # a geometry input returns a geometry
    actual = set_coordinates(point, [[1, 1]])
    assert isinstance(actual, shapely.Geometry)
    # a 0-dim array input returns a 0-dim array
    actual = set_coordinates(np.asarray(point), [[1, 1]])
    assert isinstance(actual, np.ndarray)
    assert actual.ndim == 0


@pytest.mark.parametrize("include_z", [True, False])
def test_set_coords_mixed_dimension(include_z):
    geoms = np.array([point, point_z], dtype=object)
    coords = get_coordinates(geoms, include_z=include_z)
    new_geoms = set_coordinates(geoms, coords * 2)
    if include_z:
        # preserve original dimensionality
        assert not shapely.has_z(new_geoms[0])
        assert shapely.has_z(new_geoms[1])
    else:
        # all 2D
        assert not shapely.has_z(new_geoms).any()


@pytest.mark.parametrize("include_z", [True, False])
@pytest.mark.parametrize(
    "geoms",
    [[], [empty], [None, point, None], [nested_3], [point, point_z], [line_string_z]],
)
@pytest.mark.parametrize("interleaved", [True, False])
def test_transform(geoms, include_z, interleaved):
    geoms = np.array(geoms, np.object_)
    coordinates_before = get_coordinates(geoms, include_z=include_z)
    if interleaved:
        transformation = lambda coords: coords + 1
    elif not include_z:
        transformation = lambda x, y: (x + 1, y + 1)
    else:
        transformation = lambda x, y, z: (x + 1, y + 1, z + 1)
    new_geoms = transform(
        geoms, transformation, include_z=include_z, interleaved=interleaved
    )
    assert new_geoms is not geoms
    coordinates_after = get_coordinates(new_geoms, include_z=include_z)
    assert_allclose(coordinates_before + 1, coordinates_after, equal_nan=True)


def test_transform_0dim():
    # a geometry input returns a geometry
    actual = transform(point, lambda x: x + 1)
    assert isinstance(actual, shapely.Geometry)
    # a 0-dim array input returns a 0-dim array
    actual = transform(np.asarray(point), lambda x: x + 1)
    assert isinstance(actual, np.ndarray)
    assert actual.ndim == 0


def test_transform_no_geoms():
    # a geometry input returns a geometry
    actual = transform([], lambda x: x + 1)
    assert actual.shape == (0,)


def test_transform_check_shape():
    def remove_coord(arr):
        return arr[:-1]

    with pytest.raises(ValueError):
        transform(linear_ring, remove_coord)


def test_transform_correct_coordinate_dimension():
    # ensure that new geometry is 2D with include_z=False
    geom = line_string_z
    assert shapely.get_coordinate_dimension(geom) == 3
    new_geom = transform(geom, lambda x: x + 1, include_z=False)
    assert shapely.get_coordinate_dimension(new_geom) == 2


@pytest.mark.parametrize(
    "geom",
    [
        empty_point_z,
        empty_line_string_z,
    ],
)
def test_transform_empty_preserve_z(geom):
    assert shapely.get_coordinate_dimension(geom) == 3
    new_geom = transform(geom, lambda x: x + 1, include_z=True)
    assert shapely.get_coordinate_dimension(new_geom) == 3


@pytest.mark.parametrize(
    "geom",
    [
        empty_point_z,
        empty_line_string_z,
    ],
)
def test_transform_remove_z(geom):
    assert shapely.get_coordinate_dimension(geom) == 3
    new_geom = transform(geom, lambda x: x + 1, include_z=False)
    assert shapely.get_coordinate_dimension(new_geom) == 2


@pytest.mark.parametrize(
    "geom,expected",
    [
        (line_string, 2),
        (line_string_z, 3),
    ],
)
def test_transform_auto_coordinate_dimension(geom, expected):
    new_geom = transform(geom, lambda x: x + 1, include_z=None)
    assert (shapely.get_coordinate_dimension(new_geom) == expected).all()


def test_transform_auto_coordinate_dimension_mixed():
    new_geom = transform([line_string, line_string_z], lambda x: x + 1, include_z=None)
    assert_equal(shapely.get_coordinate_dimension(new_geom), [2, 3])
    assert_equal(
        shapely.get_coordinates(line_string, include_z=False) + 1,
        shapely.get_coordinates(new_geom[0], include_z=False),
    )
    assert_equal(
        shapely.get_coordinates(line_string_z, include_z=True) + 1,
        shapely.get_coordinates(new_geom[1], include_z=True),
    )


def transform_non_interleaved(x, y, z=None):
    if z is None:
        return [x + 1, y + 2]
    else:
        return [x + 1, y + 2, z + 3]


def test_transform_auto_coordinate_dimension_mixed_interleaved():
    new_geom = transform(
        [line_string, line_string_z],
        transform_non_interleaved,
        include_z=None,
        interleaved=False,
    )
    assert_equal(shapely.get_coordinate_dimension(new_geom), [2, 3])
    assert_equal(
        shapely.get_coordinates(line_string, include_z=False) + [1, 2],
        shapely.get_coordinates(new_geom[0], include_z=False),
    )
    assert_equal(
        shapely.get_coordinates(line_string_z, include_z=True) + [1, 2, 3],
        shapely.get_coordinates(new_geom[1], include_z=True),
    )


def test_transform_deprecate_positional():
    with pytest.deprecated_call(
        match="positional argument `include_z` for `transform` is deprecated"
    ):
        transform(line_string_z, lambda x: x + 1, False)



"""Support for GEOS prepared geometry operations."""

from pickle import PicklingError

import shapely


class PreparedGeometry:
    """A geometry prepared for efficient comparison to a set of other geometries.

    Examples
    --------
    >>> from shapely.prepared import prep
    >>> from shapely.geometry import Point, Polygon
    >>> triangle = Polygon([(0.0, 0.0), (1.0, 1.0), (1.0, -1.0)])
    >>> p = prep(triangle)
    >>> p.intersects(Point(0.5, 0.5))
    True

    """

    def __init__(self, context):
        """Prepare a geometry for efficient comparison to other geometries."""
        if isinstance(context, PreparedGeometry):
            self.context = context.context
        else:
            shapely.prepare(context)
            self.context = context
        self.prepared = True

    def contains(self, other):
        """Return True if the geometry contains the other, else False."""
        return self.context.contains(other)

    def contains_properly(self, other):
        """Return True if the geometry properly contains the other, else False."""
        return self.context.contains_properly(other)

    def covers(self, other):
        """Return True if the geometry covers the other, else False."""
        return self.context.covers(other)

    def crosses(self, other):
        """Return True if the geometries cross, else False."""
        return self.context.crosses(other)

    def disjoint(self, other):
        """Return True if geometries are disjoint, else False."""
        return self.context.disjoint(other)

    def intersects(self, other):
        """Return True if geometries intersect, else False."""
        return self.context.intersects(other)

    def overlaps(self, other):
        """Return True if geometries overlap, else False."""
        return self.context.overlaps(other)

    def touches(self, other):
        """Return True if geometries touch, else False."""
        return self.context.touches(other)

    def within(self, other):
        """Return True if geometry is within the other, else False."""
        return self.context.within(other)

    def __reduce__(self):
        """Pickling is not supported."""
        raise PicklingError("Prepared geometries cannot be pickled.")


def prep(ob):
    """Create and return a prepared geometric object."""
    return PreparedGeometry(ob)



"""Proxies for libgeos, GEOS-specific exceptions, and utilities."""

import warnings

import shapely

warnings.warn(
    "The 'shapely.geos' module is deprecated, and will be removed in a future version. "
    "All attributes of 'shapely.geos' are available directly from the top-level "
    "'shapely' namespace (since shapely 2.0.0).",
    DeprecationWarning,
    stacklevel=2,
)

geos_version_string = shapely.geos_capi_version_string
geos_version = shapely.geos_version
geos_capi_version = shapely.geos_capi_version



"""Speedups for Shapely geometry operations.

.. deprecated:: 2.0
        Deprecated in Shapely 2.0, and will be removed in a future version.

"""

import warnings

__all__ = ["available", "disable", "enable", "enabled"]


available = True
enabled = True


_MSG = (
    "This function has no longer any effect, and will be removed in a "
    "future release. Starting with Shapely 2.0, equivalent speedups are "
    "always available"
)


def enable():
    """Will be removed in a future release and has no longer any effect.

    Previously, this function enabled cython-based speedups. Starting with
    Shapely 2.0, equivalent speedups are available in every installation.
    """
    warnings.warn(_MSG, FutureWarning, stacklevel=2)


def disable():
    """Will be removed in a future release and has no longer any effect.

    Previously, this function enabled cython-based speedups. Starting with
    Shapely 2.0, equivalent speedups are available in every installation.
    """
    warnings.warn(_MSG, FutureWarning, stacklevel=2)



import numpy as np
import pytest
from numpy.testing import assert_allclose

import shapely
from shapely import MultiLineString, MultiPoint, MultiPolygon
from shapely.testing import assert_geometries_equal
from shapely.tests.common import (
    empty_line_string,
    empty_line_string_m,
    empty_line_string_z,
    empty_line_string_zm,
    empty_multi_polygon_m,
    empty_multi_polygon_z,
    empty_multi_polygon_zm,
    geometry_collection,
    line_string,
    line_string_m,
    line_string_z,
    line_string_zm,
    linear_ring,
    multi_line_string,
    multi_line_string_m,
    multi_line_string_z,
    multi_line_string_zm,
    multi_point,
    multi_point_m,
    multi_point_z,
    multi_point_zm,
    multi_polygon,
    multi_polygon_m,
    multi_polygon_z,
    multi_polygon_zm,
    point,
    point_m,
    point_z,
    point_zm,
    polygon,
    polygon_m,
    polygon_z,
    polygon_zm,
)

all_types = (
    point,
    line_string,
    polygon,
    multi_point,
    multi_line_string,
    multi_polygon,
)

all_types_z = (
    point_z,
    line_string_z,
    polygon_z,
    multi_point_z,
    multi_line_string_z,
    multi_polygon_z,
)

all_types_m = (
    point_m,
    line_string_m,
    polygon_m,
    multi_point_m,
    multi_line_string_m,
    multi_polygon_m,
)

all_types_zm = (
    point_zm,
    line_string_zm,
    polygon_zm,
    multi_point_zm,
    multi_line_string_zm,
    multi_polygon_zm,
)

all_types_dims_combos = all_types + all_types_z
if shapely.geos_version >= (3, 12, 0):
    all_types_dims_combos = all_types_dims_combos + all_types_m + all_types_zm

all_types_not_supported = (
    linear_ring,
    geometry_collection,
)


@pytest.mark.parametrize("geom", all_types + all_types_z)
def test_roundtrip(geom):
    actual = shapely.from_ragged_array(*shapely.to_ragged_array([geom, geom]))
    assert_geometries_equal(actual, [geom, geom])


@pytest.mark.parametrize("include_m", [None, True, False])
@pytest.mark.parametrize("include_z", [None, True, False])
@pytest.mark.parametrize("geom", all_types_dims_combos)
def test_to_ragged_array(geom, include_z, include_m):
    _, coords, _ = shapely.to_ragged_array(
        [geom, geom], include_z=include_z, include_m=include_m
    )
    nan_dims = np.all(np.isnan(coords), axis=0).tolist()
    expected = [False, False]  # XY
    has_z = geom.has_z
    if include_z or (include_z is None and has_z):
        expected.append(not has_z)  # XYZ or XYZM
    if shapely.geos_version >= (3, 12, 0):
        has_m = geom.has_m
    else:
        has_m = False
    if include_m or (include_m is None and has_m):
        expected.append(not has_m)  # XYM or XYZM
    assert nan_dims == expected


def test_include_z_default():
    # corner cases for inferring dimensionality

    # mixed XY and XYZ -> XYZ
    _, coords, _ = shapely.to_ragged_array([line_string, line_string_z])
    assert coords.shape[1] == 3

    # only empties -> always 2D
    _, coords, _ = shapely.to_ragged_array([empty_line_string])
    assert coords.shape[1] == 2
    _, coords, _ = shapely.to_ragged_array([empty_line_string_z])
    assert coords.shape[1] == 2
    # empty collection -> GEOS indicates 2D
    _, coords, _ = shapely.to_ragged_array([empty_multi_polygon_z])
    assert coords.shape[1] == 2


@pytest.mark.skipif(shapely.geos_version < (3, 12, 0), reason="GEOS < 3.12")
def test_include_m_default():
    # a few other corner cases for inferring dimensionality

    # mixed XY and XYM -> XYM
    _, coords, _ = shapely.to_ragged_array([line_string, line_string_m])
    assert coords.shape[1] == 3

    # mixed XY, XYM, and XYZM -> XYZM
    _, coords, _ = shapely.to_ragged_array([line_string, line_string_m, line_string_zm])
    assert coords.shape[1] == 4

    # only empties -> always 2D
    _, coords, _ = shapely.to_ragged_array([empty_line_string_m])
    assert coords.shape[1] == 2
    _, coords, _ = shapely.to_ragged_array([empty_line_string_zm])
    assert coords.shape[1] == 2
    # empty collection -> GEOS indicates 2D
    _, coords, _ = shapely.to_ragged_array([empty_multi_polygon_m])
    assert coords.shape[1] == 2
    _, coords, _ = shapely.to_ragged_array([empty_multi_polygon_zm])
    assert coords.shape[1] == 2


@pytest.mark.parametrize("geom", all_types)
def test_read_only_arrays(geom):
    # https://github.com/shapely/shapely/pull/1744
    typ, coords, offsets = shapely.to_ragged_array([geom, geom])
    coords.flags.writeable = False
    for arr in offsets:
        arr.flags.writeable = False
    result = shapely.from_ragged_array(typ, coords, offsets)
    assert_geometries_equal(result, [geom, geom])


@pytest.mark.parametrize("geom", all_types_not_supported)
def test_raise_geometry_type(geom):
    with pytest.raises(ValueError):
        shapely.to_ragged_array([geom, geom])


def test_points():
    arr = shapely.from_wkt(
        [
            "POINT (0 0)",
            "POINT (1 1)",
            "POINT EMPTY",
            "POINT EMPTY",
            "POINT (4 4)",
            None,
            "POINT EMPTY",
        ]
    )
    typ, result, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [0, 0],
            [1, 1],
            [np.nan, np.nan],
            [np.nan, np.nan],
            [4, 4],
            [np.nan, np.nan],
            [np.nan, np.nan],
        ]
    )
    assert typ == shapely.GeometryType.POINT
    assert len(result) == len(arr)
    assert_allclose(result, expected)
    assert len(offsets) == 0

    geoms = shapely.from_ragged_array(typ, result)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("POINT EMPTY")
    assert_geometries_equal(geoms, arr)


def test_linestrings():
    arr = shapely.from_wkt(
        [
            "LINESTRING (30 10, 10 30, 40 40)",
            "LINESTRING (40 40, 30 30, 40 20, 30 10)",
            "LINESTRING EMPTY",
            "LINESTRING EMPTY",
            "LINESTRING (10 10, 20 20, 10 40)",
            None,
            "LINESTRING EMPTY",
        ]
    )
    typ, coords, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [30.0, 10.0],
            [10.0, 30.0],
            [40.0, 40.0],
            [40.0, 40.0],
            [30.0, 30.0],
            [40.0, 20.0],
            [30.0, 10.0],
            [10.0, 10.0],
            [20.0, 20.0],
            [10.0, 40.0],
        ]
    )
    expected_offsets = np.array([0, 3, 7, 7, 7, 10, 10, 10], dtype="int32")
    assert typ == shapely.GeometryType.LINESTRING
    assert_allclose(coords, expected)
    assert len(offsets) == 1
    assert offsets[0].dtype == np.int32
    assert_allclose(offsets[0], expected_offsets)

    result = shapely.from_ragged_array(typ, coords, offsets)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("LINESTRING EMPTY")
    assert_geometries_equal(result, arr)

    # sliced
    offsets_sliced = (offsets[0][1:],)
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[1:])

    offsets_sliced = (offsets[0][:-1],)
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[:-1])


def test_polygons():
    arr = shapely.from_wkt(
        [
            "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))",
            "POLYGON ((35 10, 45 45, 15 40, 10 20, 35 10), (20 30, 35 35, 30 20, 20 30))",  # noqa: E501
            "POLYGON EMPTY",
            "POLYGON EMPTY",
            "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))",
            None,
            "POLYGON EMPTY",
        ]
    )
    typ, coords, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [30.0, 10.0],
            [40.0, 40.0],
            [20.0, 40.0],
            [10.0, 20.0],
            [30.0, 10.0],
            [35.0, 10.0],
            [45.0, 45.0],
            [15.0, 40.0],
            [10.0, 20.0],
            [35.0, 10.0],
            [20.0, 30.0],
            [35.0, 35.0],
            [30.0, 20.0],
            [20.0, 30.0],
            [30.0, 10.0],
            [40.0, 40.0],
            [20.0, 40.0],
            [10.0, 20.0],
            [30.0, 10.0],
        ]
    )
    expected_offsets1 = np.array([0, 5, 10, 14, 19])
    expected_offsets2 = np.array([0, 1, 3, 3, 3, 4, 4, 4])

    assert typ == shapely.GeometryType.POLYGON
    assert_allclose(coords, expected)
    assert len(offsets) == 2
    assert offsets[0].dtype == np.int32
    assert offsets[1].dtype == np.int32
    assert_allclose(offsets[0], expected_offsets1)
    assert_allclose(offsets[1], expected_offsets2)

    result = shapely.from_ragged_array(typ, coords, offsets)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("POLYGON EMPTY")
    assert_geometries_equal(result, arr)

    # sliced:
    # - indices into the coordinate array for the whole buffer
    # - indices into the ring array for *just* the sliced part
    offsets_sliced = (offsets[0], offsets[1][1:])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[1:])

    offsets_sliced = (offsets[0], offsets[1][:-1])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[:-1])


def test_multipoints():
    arr = shapely.from_wkt(
        [
            "MULTIPOINT (10 40, 40 30, 20 20, 30 10)",
            "MULTIPOINT (30 10)",
            "MULTIPOINT EMPTY",
            "MULTIPOINT EMPTY",
            "MULTIPOINT (30 10, 10 30, 40 40)",
            None,
            "MULTIPOINT EMPTY",
        ]
    )
    typ, coords, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [10.0, 40.0],
            [40.0, 30.0],
            [20.0, 20.0],
            [30.0, 10.0],
            [30.0, 10.0],
            [30.0, 10.0],
            [10.0, 30.0],
            [40.0, 40.0],
        ]
    )
    expected_offsets = np.array([0, 4, 5, 5, 5, 8, 8, 8])

    assert typ == shapely.GeometryType.MULTIPOINT
    assert_allclose(coords, expected)
    assert len(offsets) == 1
    assert offsets[0].dtype == np.int32
    assert_allclose(offsets[0], expected_offsets)

    result = shapely.from_ragged_array(typ, coords, offsets)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("MULTIPOINT EMPTY")
    assert_geometries_equal(result, arr)

    # sliced:
    offsets_sliced = (offsets[0][1:],)
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[1:])

    offsets_sliced = (offsets[0][:-1],)
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[:-1])


def test_multilinestrings():
    arr = shapely.from_wkt(
        [
            "MULTILINESTRING ((30 10, 10 30, 40 40))",
            "MULTILINESTRING ((10 10, 20 20, 10 40), (40 40, 30 30, 40 20, 30 10))",
            "MULTILINESTRING EMPTY",
            "MULTILINESTRING EMPTY",
            "MULTILINESTRING ((35 10, 45 45), (15 40, 10 20), (30 10, 10 30, 40 40))",
            None,
            "MULTILINESTRING EMPTY",
        ]
    )
    typ, coords, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [30.0, 10.0],
            [10.0, 30.0],
            [40.0, 40.0],
            [10.0, 10.0],
            [20.0, 20.0],
            [10.0, 40.0],
            [40.0, 40.0],
            [30.0, 30.0],
            [40.0, 20.0],
            [30.0, 10.0],
            [35.0, 10.0],
            [45.0, 45.0],
            [15.0, 40.0],
            [10.0, 20.0],
            [30.0, 10.0],
            [10.0, 30.0],
            [40.0, 40.0],
        ]
    )
    expected_offsets1 = np.array([0, 3, 6, 10, 12, 14, 17])
    expected_offsets2 = np.array([0, 1, 3, 3, 3, 6, 6, 6])

    assert typ == shapely.GeometryType.MULTILINESTRING
    assert_allclose(coords, expected)
    assert len(offsets) == 2
    assert offsets[0].dtype == np.int32
    assert offsets[1].dtype == np.int32
    assert_allclose(offsets[0], expected_offsets1)
    assert_allclose(offsets[1], expected_offsets2)

    result = shapely.from_ragged_array(typ, coords, offsets)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("MULTILINESTRING EMPTY")
    assert_geometries_equal(result, arr)

    # sliced:
    # - indices into the coordinate array for the whole buffer
    # - indices into the line parts for *just* the sliced part
    offsets_sliced = (offsets[0], offsets[1][1:])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[1:])

    offsets_sliced = (offsets[0], offsets[1][:-1])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[:-1])


def test_multipolygons():
    arr = shapely.from_wkt(
        [
            "MULTIPOLYGON (((35 10, 45 45, 15 40, 10 20, 35 10), (20 30, 35 35, 30 20, 20 30)))",  # noqa: E501
            "MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)), ((20 35, 10 30, 10 10, 30 5, 45 20, 20 35), (30 20, 20 15, 20 25, 30 20)))",  # noqa: E501
            "MULTIPOLYGON EMPTY",
            "MULTIPOLYGON EMPTY",
            "MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)))",
            None,
            "MULTIPOLYGON EMPTY",
        ]
    )
    typ, coords, offsets = shapely.to_ragged_array(arr)
    expected = np.array(
        [
            [35.0, 10.0],
            [45.0, 45.0],
            [15.0, 40.0],
            [10.0, 20.0],
            [35.0, 10.0],
            [20.0, 30.0],
            [35.0, 35.0],
            [30.0, 20.0],
            [20.0, 30.0],
            [40.0, 40.0],
            [20.0, 45.0],
            [45.0, 30.0],
            [40.0, 40.0],
            [20.0, 35.0],
            [10.0, 30.0],
            [10.0, 10.0],
            [30.0, 5.0],
            [45.0, 20.0],
            [20.0, 35.0],
            [30.0, 20.0],
            [20.0, 15.0],
            [20.0, 25.0],
            [30.0, 20.0],
            [40.0, 40.0],
            [20.0, 45.0],
            [45.0, 30.0],
            [40.0, 40.0],
        ]
    )
    expected_offsets1 = np.array([0, 5, 9, 13, 19, 23, 27])
    expected_offsets2 = np.array([0, 2, 3, 5, 6])
    expected_offsets3 = np.array([0, 1, 3, 3, 3, 4, 4, 4])

    assert typ == shapely.GeometryType.MULTIPOLYGON
    assert_allclose(coords, expected)
    assert len(offsets) == 3
    assert offsets[0].dtype == np.int32
    assert offsets[1].dtype == np.int32
    assert offsets[2].dtype == np.int32
    assert_allclose(offsets[0], expected_offsets1)
    assert_allclose(offsets[1], expected_offsets2)
    assert_allclose(offsets[2], expected_offsets3)

    result = shapely.from_ragged_array(typ, coords, offsets)
    # in a roundtrip, missing geometries come back as empty
    arr[-2] = shapely.from_wkt("MULTIPOLYGON EMPTY")
    assert_geometries_equal(result, arr)

    # sliced:
    offsets_sliced = (offsets[0], offsets[1], offsets[2][1:])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[1:])

    offsets_sliced = (offsets[0], offsets[1], offsets[2][:-3])
    result = shapely.from_ragged_array(typ, coords, offsets_sliced)
    assert_geometries_equal(result, arr[:-3])
    print(result)


def test_mixture_point_multipoint():
    typ, coords, offsets = shapely.to_ragged_array([point, multi_point])
    assert typ == shapely.GeometryType.MULTIPOINT
    result = shapely.from_ragged_array(typ, coords, offsets)
    expected = np.array([MultiPoint([point]), multi_point])
    assert_geometries_equal(result, expected)


def test_mixture_linestring_multilinestring():
    typ, coords, offsets = shapely.to_ragged_array([line_string, multi_line_string])
    assert typ == shapely.GeometryType.MULTILINESTRING
    result = shapely.from_ragged_array(typ, coords, offsets)
    expected = np.array([MultiLineString([line_string]), multi_line_string])
    assert_geometries_equal(result, expected)


def test_mixture_polygon_multipolygon():
    typ, coords, offsets = shapely.to_ragged_array([polygon, multi_polygon])
    assert typ == shapely.GeometryType.MULTIPOLYGON
    result = shapely.from_ragged_array(typ, coords, offsets)
    expected = np.array([MultiPolygon([polygon]), multi_polygon])
    assert_geometries_equal(result, expected)


def test_from_ragged_incorrect_rings_short():
    # too few coordinates for linearring
    coords = np.array([[0, 0], [1, 1]], dtype="float64")
    offsets1 = np.array([0, 2])
    offsets2 = np.array([0, 1])
    offsets3 = np.array([0, 1])

    with pytest.raises(
        ValueError, match="A linearring requires at least 4 coordinates"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )

    with pytest.raises(
        ValueError, match="A linearring requires at least 4 coordinates"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.POLYGON, coords, (offsets1, offsets2)
        )


def test_from_ragged_incorrect_rings_unclosed():
    # NaNs cause the ring to be unclosed
    coords = np.full((4, 2), np.nan)
    offsets1 = np.array([0, 4])
    offsets2 = np.array([0, 1])
    offsets3 = np.array([0, 1])

    with pytest.raises(
        shapely.GEOSException,
        match="Points of LinearRing do not form a closed linestring",
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )

    with pytest.raises(
        shapely.GEOSException,
        match="Points of LinearRing do not form a closed linestring",
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.POLYGON, coords, (offsets1, offsets2)
        )


def test_from_ragged_wrong_offsets():
    with pytest.raises(ValueError, match="'offsets' must be provided"):
        shapely.from_ragged_array(
            shapely.GeometryType.LINESTRING, np.array([[0, 0], [0, 1]])
        )

    with pytest.raises(ValueError, match="'offsets' should not be provided"):
        shapely.from_ragged_array(
            shapely.GeometryType.POINT,
            np.array([[0, 0], [0, 1]]),
            offsets=(np.array([0, 1]),),
        )


def test_from_ragged_crash_2284():
    # caused segfault in shapely 2.1.0
    # https://github.com/shapely/shapely/discussions/2284

    # one of the geometries has more rings than the total number of geometries
    coords = np.random.default_rng().random(120).reshape((60, 2))
    offsets1 = np.array([0, 10, 20, 30, 40, 50, 60])
    offsets2 = np.array([0, 1, 5, 6])

    for _ in range(10):
        polygons = shapely.from_ragged_array(
            shapely.GeometryType.POLYGON, coords, (offsets1, offsets2)
        )
        # just ensure it didn't crash
        assert len(polygons) == 3

    offsets3 = np.array([0, 3])
    for _ in range(10):
        polygons = shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )
        # just ensure it didn't crash
        assert len(polygons) == 1


def test_from_ragged_wrong_offsets_values():
    # caused segfault in shapely 2.1.0

    # outer offsets indicates more rings than the shape of the ring offsets
    coords = np.random.default_rng().random(70).reshape((35, 2))
    offsets1 = np.array([0, 10, 20], dtype=np.uint32)
    offsets2 = np.array([0, 1, 5], dtype=np.uint32)
    offsets3 = np.array([0, 2])

    with pytest.raises(
        ValueError, match="Number of rings indicated by the geometry offsets"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.POLYGON, coords, (offsets1, offsets2)
        )

    with pytest.raises(
        ValueError, match="Number of rings indicated by the part offsets"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )

    # inner offsets indicating more coordinats that the shape of the coordinates
    coords = np.random.default_rng().random(70).reshape((35, 2))
    offsets1 = np.array([0, 10, 40], dtype=np.uint32)
    offsets2 = np.array([0, 1, 2], dtype=np.uint32)

    with pytest.raises(
        ValueError, match="Number of coordinates indicated by the linear offsets"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.POLYGON, coords, (offsets1, offsets2)
        )

    with pytest.raises(
        ValueError, match="Number of coordinates indicated by the linear offsets"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )

    # outer multipolygon offsets indicating too many parts
    offsets3 = np.array([0, 3])
    with pytest.raises(
        ValueError, match="Number of geometry parts indicated by the geometry offsets"
    ):
        shapely.from_ragged_array(
            shapely.GeometryType.MULTIPOLYGON, coords, (offsets1, offsets2, offsets3)
        )



"""Manipulation and analysis of geometric objects in the Cartesian plane."""

from shapely.lib import GEOSException
from shapely.lib import Geometry
from shapely.lib import geos_version, geos_version_string
from shapely.lib import geos_capi_version, geos_capi_version_string
from shapely.errors import setup_signal_checks
from shapely._geometry import *
from shapely.creation import *
from shapely.constructive import *
from shapely.predicates import *
from shapely.measurement import *
from shapely.set_operations import *
from shapely.linear import *
from shapely.coordinates import *
from shapely.strtree import *
from shapely.io import *
from shapely._coverage import *

# Submodule always needs to be imported to ensure Geometry subclasses are registered
from shapely.geometry import (
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
    GeometryCollection,
    LinearRing,
)

from shapely import _version

__version__ = _version.get_versions()["version"]

setup_signal_checks()



import unittest
from math import pi

import numpy as np
import pytest

from shapely import affinity
from shapely.geometry import Point
from shapely.wkt import loads as load_wkt


class AffineTestCase(unittest.TestCase):
    def test_affine_params(self):
        g = load_wkt("LINESTRING(2.4 4.1, 2.4 3, 3 3)")

        with pytest.raises(TypeError):
            affinity.affine_transform(g, None)

        with pytest.raises(ValueError):
            affinity.affine_transform(g, [1, 2, 3, 4, 5, 6, 7, 8, 9])

        with pytest.raises(AttributeError):
            affinity.affine_transform(None, [1, 2, 3, 4, 5, 6])

    def test_affine_geom_types(self):
        # identity matrices, which should result with no transformation
        matrix2d = (1, 0, 0, 1, 0, 0)
        matrix3d = (1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0)

        # empty in, empty out
        empty2d = load_wkt("MULTIPOLYGON EMPTY")
        assert affinity.affine_transform(empty2d, matrix2d).is_empty

        def test_geom(g2, g3=None):
            assert not g2.has_z
            a2 = affinity.affine_transform(g2, matrix2d)
            assert not a2.has_z
            assert g2.equals(a2)
            if g3 is not None:
                assert g3.has_z
                a3 = affinity.affine_transform(g3, matrix3d)
                assert a3.has_z
                assert g3.equals(a3)

        pt2d = load_wkt("POINT(12.3 45.6)")
        pt3d = load_wkt("POINT(12.3 45.6 7.89)")
        test_geom(pt2d, pt3d)
        ls2d = load_wkt("LINESTRING(0.9 3.4, 0.7 2, 2.5 2.7)")
        ls3d = load_wkt("LINESTRING(0.9 3.4 3.3, 0.7 2 2.3, 2.5 2.7 5.5)")
        test_geom(ls2d, ls3d)
        lr2d = load_wkt("LINEARRING(0.9 3.4, 0.7 2, 2.5 2.7, 0.9 3.4)")
        lr3d = load_wkt("LINEARRING(0.9 3.4 3.3, 0.7 2 2.3, 2.5 2.7 5.5, 0.9 3.4 3.3)")
        test_geom(lr2d, lr3d)
        test_geom(
            load_wkt(
                "POLYGON((0.9 2.3, 0.5 1.1, 2.4 0.8, 0.9 2.3), "
                "(1.1 1.7, 0.9 1.3, 1.4 1.2, 1.1 1.7), "
                "(1.6 1.3, 1.7 1, 1.9 1.1, 1.6 1.3))"
            )
        )
        test_geom(
            load_wkt("MULTIPOINT ((-300 300), (700 300), (-800 -1100), (200 -300))")
        )
        test_geom(
            load_wkt(
                "MULTILINESTRING((0 0, -0.7 -0.7, 0.6 -1), (-0.5 0.5, 0.7 0.6, 0 -0.6))"
            )
        )
        test_geom(
            load_wkt(
                "MULTIPOLYGON(((900 4300, -1100 -400, 900 -800, 900 4300)), "
                "((1200 4300, 2300 4400, 1900 1000, 1200 4300)))"
            )
        )
        test_geom(
            load_wkt(
                "GEOMETRYCOLLECTION(POINT(20 70),"
                " POLYGON((60 70, 13 35, 60 -30, 60 70)),"
                " LINESTRING(60 70, 50 100, 80 100))"
            )
        )

    def test_affine_2d(self):
        g = load_wkt("LINESTRING(2.4 4.1, 2.4 3, 3 3)")
        # custom scale and translate
        expected2d = load_wkt("LINESTRING(-0.2 14.35, -0.2 11.6, 1 11.6)")
        matrix2d = (2, 0, 0, 2.5, -5, 4.1)
        a2 = affinity.affine_transform(g, matrix2d)
        assert a2.equals_exact(expected2d, 1e-6)
        assert not a2.has_z
        # Make sure a 3D matrix does not make a 3D shape from a 2D input
        matrix3d = (2, 0, 0, 0, 2.5, 0, 0, 0, 10, -5, 4.1, 100)
        a3 = affinity.affine_transform(g, matrix3d)
        assert a3.equals_exact(expected2d, 1e-6)
        assert not a3.has_z

    def test_affine_3d(self):
        g2 = load_wkt("LINESTRING(2.4 4.1, 2.4 3, 3 3)")
        g3 = load_wkt("LINESTRING(2.4 4.1 100.2, 2.4 3 132.8, 3 3 128.6)")
        # custom scale and translate
        matrix2d = (2, 0, 0, 2.5, -5, 4.1)
        matrix3d = (2, 0, 0, 0, 2.5, 0, 0, 0, 0.3048, -5, 4.1, 100)
        # Combinations of 2D and 3D geometries and matrices
        a22 = affinity.affine_transform(g2, matrix2d)
        a23 = affinity.affine_transform(g2, matrix3d)
        a32 = affinity.affine_transform(g3, matrix2d)
        a33 = affinity.affine_transform(g3, matrix3d)
        # Check dimensions
        assert not a22.has_z
        assert not a23.has_z
        assert a32.has_z
        assert a33.has_z

        # 2D equality checks
        expected2d = load_wkt("LINESTRING(-0.2 14.35, -0.2 11.6, 1 11.6)")
        expected3d = load_wkt(
            "LINESTRING(-0.2 14.35 130.54096, -0.2 11.6 140.47744, 1 11.6 139.19728)"
        )
        expected32 = load_wkt(
            "LINESTRING(-0.2 14.35 100.2, -0.2 11.6 132.8, 1 11.6 128.6)"
        )
        assert a22.equals_exact(expected2d, 1e-6)
        assert a23.equals_exact(expected2d, 1e-6)
        # Do explicit 3D check of coordinate values
        for a, e in zip(a32.coords, expected32.coords):
            for ap, ep in zip(a, e):
                self.assertAlmostEqual(ap, ep)
        for a, e in zip(a33.coords, expected3d.coords):
            for ap, ep in zip(a, e):
                self.assertAlmostEqual(ap, ep)


class TransformOpsTestCase(unittest.TestCase):
    def test_rotate(self):
        ls = load_wkt("LINESTRING(240 400, 240 300, 300 300)")
        # counter-clockwise degrees
        rls = affinity.rotate(ls, 90)
        els = load_wkt("LINESTRING(220 320, 320 320, 320 380)")
        assert rls.equals(els)
        # retest with named parameters for the same result
        rls = affinity.rotate(geom=ls, angle=90, origin="center")
        assert rls.equals(els)
        # clockwise radians
        rls = affinity.rotate(ls, -pi / 2, use_radians=True)
        els = load_wkt("LINESTRING(320 380, 220 380, 220 320)")
        assert rls.equals(els)
        ## other `origin` parameters
        # around the centroid
        rls = affinity.rotate(ls, 90, origin="centroid")
        els = load_wkt("LINESTRING(182.5 320, 282.5 320, 282.5 380)")
        assert rls.equals(els)
        # around the second coordinate tuple
        rls = affinity.rotate(ls, 90, origin=ls.coords[1])
        els = load_wkt("LINESTRING(140 300, 240 300, 240 360)")
        assert rls.equals(els)
        # around the absolute Point of origin
        rls = affinity.rotate(ls, 90, origin=Point(0, 0))
        els = load_wkt("LINESTRING(-400 240, -300 240, -300 300)")
        assert rls.equals(els)

    def test_rotate_empty(self):
        rls = affinity.rotate(load_wkt("LINESTRING EMPTY"), 90)
        els = load_wkt("LINESTRING EMPTY")
        assert rls.equals(els)

    def test_rotate_angle_array(self):
        ls = load_wkt("LINESTRING(240 400, 240 300, 300 300)")
        els = load_wkt("LINESTRING(220 320, 320 320, 320 380)")
        # check with degrees
        theta = np.array(90.0)
        rls = affinity.rotate(ls, theta)
        assert theta.item() == 90.0
        assert rls.equals(els)
        # check with radians
        theta = np.array(pi / 2)
        rls = affinity.rotate(ls, theta, use_radians=True)
        assert theta.item() == pi / 2
        assert rls.equals(els)

    def test_scale(self):
        ls = load_wkt("LINESTRING(240 400 10, 240 300 30, 300 300 20)")
        # test defaults of 1.0
        sls = affinity.scale(ls)
        assert sls.equals(ls)
        # different scaling in different dimensions
        sls = affinity.scale(ls, 2, 3, 0.5)
        els = load_wkt("LINESTRING(210 500 5, 210 200 15, 330 200 10)")
        assert sls.equals(els)
        # Do explicit 3D check of coordinate values
        for a, b in zip(sls.coords, els.coords):
            for ap, bp in zip(a, b):
                self.assertEqual(ap, bp)
        # retest with named parameters for the same result
        sls = affinity.scale(geom=ls, xfact=2, yfact=3, zfact=0.5, origin="center")
        assert sls.equals(els)
        ## other `origin` parameters
        # around the centroid
        sls = affinity.scale(ls, 2, 3, 0.5, origin="centroid")
        els = load_wkt("LINESTRING(228.75 537.5, 228.75 237.5, 348.75 237.5)")
        assert sls.equals(els)
        # around the second coordinate tuple
        sls = affinity.scale(ls, 2, 3, 0.5, origin=ls.coords[1])
        els = load_wkt("LINESTRING(240 600, 240 300, 360 300)")
        assert sls.equals(els)
        # around some other 3D Point of origin
        sls = affinity.scale(ls, 2, 3, 0.5, origin=Point(100, 200, 1000))
        els = load_wkt("LINESTRING(380 800 505, 380 500 515, 500 500 510)")
        assert sls.equals(els)
        # Do explicit 3D check of coordinate values
        for a, b in zip(sls.coords, els.coords):
            for ap, bp in zip(a, b):
                assert ap == bp

    def test_scale_empty(self):
        sls = affinity.scale(load_wkt("LINESTRING EMPTY"))
        els = load_wkt("LINESTRING EMPTY")
        assert sls.equals(els)

    def test_skew(self):
        ls = load_wkt("LINESTRING(240 400 10, 240 300 30, 300 300 20)")
        # test default shear angles of 0.0
        sls = affinity.skew(ls)
        assert sls.equals(ls)
        # different shearing in x- and y-directions
        sls = affinity.skew(ls, 15, -30)
        els = load_wkt(
            "LINESTRING (253.39745962155615 417.3205080756888, "
            "226.60254037844385 317.3205080756888, "
            "286.60254037844385 282.67949192431126)"
        )
        assert sls.equals_exact(els, 1e-6)
        # retest with radians for the same result
        sls = affinity.skew(ls, pi / 12, -pi / 6, use_radians=True)
        assert sls.equals_exact(els, 1e-6)
        # retest with named parameters for the same result
        sls = affinity.skew(geom=ls, xs=15, ys=-30, origin="center", use_radians=False)
        assert sls.equals_exact(els, 1e-6)
        ## other `origin` parameters
        # around the centroid
        sls = affinity.skew(ls, 15, -30, origin="centroid")
        els = load_wkt(
            "LINESTRING(258.42150697963973 406.49519052838332, "
            "231.6265877365273980 306.4951905283833185, "
            "291.6265877365274264 271.8541743770057337)"
        )
        assert sls.equals_exact(els, 1e-6)
        # around the second coordinate tuple
        sls = affinity.skew(ls, 15, -30, origin=ls.coords[1])
        els = load_wkt(
            "LINESTRING(266.7949192431123038 400, 240 300, 300 265.3589838486224153)"
        )
        assert sls.equals_exact(els, 1e-6)
        # around the absolute Point of origin
        sls = affinity.skew(ls, 15, -30, origin=Point(0, 0))
        els = load_wkt(
            "LINESTRING(347.179676972449101 261.435935394489832, "
            "320.3847577293367976 161.4359353944898317, "
            "380.3847577293367976 126.7949192431122754)"
        )
        assert sls.equals_exact(els, 1e-6)

    def test_skew_empty(self):
        sls = affinity.skew(load_wkt("LINESTRING EMPTY"))
        els = load_wkt("LINESTRING EMPTY")
        assert sls.equals(els)

    def test_skew_xs_ys_array(self):
        ls = load_wkt("LINESTRING(240 400 10, 240 300 30, 300 300 20)")
        els = load_wkt(
            "LINESTRING (253.39745962155615 417.3205080756888, "
            "226.60254037844385 317.3205080756888, "
            "286.60254037844385 282.67949192431126)"
        )
        # check with degrees
        xs_ys = np.array([15.0, -30.0])
        sls = affinity.skew(ls, xs_ys[0, ...], xs_ys[1, ...])
        assert xs_ys[0] == 15.0
        assert xs_ys[1] == -30.0
        assert sls.equals_exact(els, 1e-6)
        # check with radians
        xs_ys = np.array([pi / 12, -pi / 6])
        sls = affinity.skew(ls, xs_ys[0, ...], xs_ys[1, ...], use_radians=True)
        assert xs_ys[0] == pi / 12
        assert xs_ys[1] == -pi / 6
        assert sls.equals_exact(els, 1e-6)

    def test_translate(self):
        ls = load_wkt("LINESTRING(240 400 10, 240 300 30, 300 300 20)")
        # test default offset of 0.0
        tls = affinity.translate(ls)
        assert tls.equals(ls)
        # test all offsets
        tls = affinity.translate(ls, 100, 400, -10)
        els = load_wkt("LINESTRING(340 800 0, 340 700 20, 400 700 10)")
        assert tls.equals(els)
        # Do explicit 3D check of coordinate values
        for a, b in zip(tls.coords, els.coords):
            for ap, bp in zip(a, b):
                assert ap == bp
        # retest with named parameters for the same result
        tls = affinity.translate(geom=ls, xoff=100, yoff=400, zoff=-10)
        assert tls.equals(els)

    def test_translate_empty(self):
        tls = affinity.translate(load_wkt("LINESTRING EMPTY"))
        els = load_wkt("LINESTRING EMPTY")
        self.assertTrue(tls.equals(els))
        assert tls.equals(els)
